package org.xbib.logging.ext;

import java.util.concurrent.ConcurrentMap;
import org.xbib.logging.LogContext;
import org.xbib.logging.LogContextSelector;
import org.xbib.logging.util.CopyOnWriteMap;
import static java.lang.Thread.currentThread;

/**
 * A log context selector which chooses a log context based on the thread context classloader.
 */
public final class ContextClassLoaderLogContextSelector implements LogContextSelector {
    /**
     * Construct a new instance. If no matching log context is found, the provided default selector is consulted.
     *
     * @param defaultSelector the selector to consult if no matching log context is found
     */
    public ContextClassLoaderLogContextSelector(final LogContextSelector defaultSelector) {
        this.defaultSelector = defaultSelector;
    }

    /**
     * Construct a new instance. If no matching log context is found, the system context is used.
     */
    public ContextClassLoaderLogContextSelector() {
        this(LogContext.DEFAULT_LOG_CONTEXT_SELECTOR);
    }

    private final LogContextSelector defaultSelector;

    private final ConcurrentMap<ClassLoader, LogContext> contextMap = new CopyOnWriteMap<>();

    @Override
    public LogContext getLogContext() {
        ClassLoader cl = currentThread().getContextClassLoader();
        if (cl != null) {
            final LogContext mappedContext = contextMap.get(cl);
            if (mappedContext != null) {
                return mappedContext;
            }
        }
        return defaultSelector.getLogContext();
    }

    /**
     * Register a class loader with a log context. This method requires the {@code registerLogContext}
     * {@link RuntimePermission}.
     *
     * @param classLoader the classloader
     * @param logContext  the log context
     * @throws IllegalArgumentException if the classloader is already associated with a log context
     */
    public void registerLogContext(ClassLoader classLoader, LogContext logContext) throws IllegalArgumentException {
        if (contextMap.putIfAbsent(classLoader, logContext) != null) {
            throw new IllegalArgumentException(
                    "ClassLoader instance is already registered to a log context (" + classLoader + ")");
        }
    }

    /**
     * Unregister a class loader/log context association. This method requires the {@code unregisterLogContext}
     * {@link RuntimePermission}.
     *
     * @param classLoader the classloader
     * @param logContext  the log context
     * @return {@code true} if the association exists and was removed, {@code false} otherwise
     */
    public boolean unregisterLogContext(ClassLoader classLoader, LogContext logContext) {
        return contextMap.remove(classLoader, logContext);
    }
}
